#include "../std.h"
#include <iostream>

#include "../misc/condition.h"
#include "../misc/log_time.h"
#include "../misc/checksum.h"
#include "../misc/inttostring.h"
#include "../https/se/ns.nsmap"
#include "../https/client/client.h"
#include "../misc/time_utils.h"
#include "../../userinterface/CertInfo.h"
#include "databufferpar.h"
#include "datapoint.h"
#include "datahandle.h"

ns__fileinfo::ns__fileinfo() {
}
ns__fileinfo::~ns__fileinfo() {
}

class httpg_info_t;
static int read_callback(unsigned long long offset,unsigned long long size,char* buf,void* arg);
static void* read_thread(void *arg);
static void* write_thread(void *arg);

typedef struct {
  httpg_info_t* stat;
  unsigned long long offset;
//  unsigned long long end;
  char* buffer;
  int buffer_h;
  unsigned int buffer_size;
  unsigned int buffer_used;
  pthread_t thr;
  HTTP_Client* s;
} httpg_state_t;

class httpg_info_t {
 friend int read_callback(unsigned long long offset,unsigned long long size,char* buf,void* arg);
 friend void* read_thread(void *arg);
 friend void* write_thread(void *arg);
 friend class DataHandle;
 private:
  int streams;
  unsigned int chunk_size;
  unsigned long long int requested_size;
  int threads;
  bool have_threads;
  CondSimple lock;
  DataBufferPar* buffer;
  const char* url;
  httpg_state_t* channels;
  bool cancel;
  unsigned long long int content_size;
  bool have_content_size;
  DataHandle* handle;
  DataPoint* point;
 public:
  httpg_info_t(void);
  ~httpg_info_t(void);
};

httpg_info_t::httpg_info_t(void) {
  url=NULL;
  buffer=NULL;
  threads=0; have_threads=false;
  channels=NULL;
  cancel=false;
  have_content_size=false;
  content_size=0;
}

httpg_info_t::~httpg_info_t(void) {
  if(channels) { free(channels); channels=NULL; };
}


static int read_callback(unsigned long long offset,unsigned long long size,char* buf,void* arg) {
  httpg_state_t* tstat = (httpg_state_t*)arg;
  if(tstat->buffer == NULL) return -1;
  if(tstat->buffer_used == 0) {
    tstat->offset=offset;
  };
  // move data from buf to buffer
  for(;;) {
    if(size==0) break;
    unsigned long long l = (tstat->buffer_size - tstat->buffer_used);
    if(l>size) l=size; 
    memcpy(tstat->buffer+tstat->buffer_used,buf,l);
    tstat->buffer_used+=size;
    size-=l; offset+=l; 
    if(tstat->buffer_used == tstat->buffer_size) { // full buffer
      if(!tstat->stat->buffer->is_read(tstat->buffer_h,
			            tstat->buffer_used,tstat->offset)) {
        tstat->buffer=NULL; tstat->buffer_h=-1;
        tstat->stat->buffer->error_read(true);
        return -1;

      };
      tstat->offset+=tstat->buffer_used;
      // new buffer needed
      if(!tstat->stat->buffer->for_read(tstat->buffer_h,
			                   tstat->buffer_size,true)) {
        tstat->buffer=NULL; tstat->buffer_h=-1;
        return -1;
      };
      tstat->buffer=(*(tstat->stat->buffer))[tstat->buffer_h];
      tstat->buffer_used=0;
    };
  };
  return 0;
}

static void* read_thread(void *arg) {
  httpg_info_t* istat = (httpg_info_t*)arg;
  istat->lock.block();
  if(istat->cancel) { istat->lock.unblock(); return NULL; };
  httpg_state_t* tstat = istat->channels+istat->threads;
  istat->threads++; istat->have_threads=true;
  HTTP_Client s(istat->url);
  tstat->s=&s;
  istat->lock.unblock();
  tstat->stat=istat;
  bool failed = true;
  if(s) for(;;) {
    // get first buffer
    if(!istat->buffer->for_read(tstat->buffer_h,tstat->buffer_size,true)) break;
    tstat->buffer=(*(istat->buffer))[tstat->buffer_h];
    tstat->buffer_used=0;
    // connect
cerr<<"read_thread: calling connect"<<endl;
    if(s.connect() != 0) break;
    // acquire chunk
    istat->lock.block();
    tstat->offset = istat->requested_size;
    // check for eof and cancel
    if((istat->cancel) || 
       (istat->have_content_size && (tstat->offset >= istat->content_size))) {
cerr<<"read_thread: exiting due to eof or cancel: "<<tstat->offset<<" - "<<istat->content_size<<endl;
      istat->lock.unblock();
      istat->buffer->is_read(tstat->buffer_h,0,tstat->offset);
      if(!istat->cancel) failed=false;
      break;
    };
    istat->requested_size+=istat->chunk_size;
    istat->lock.unblock();
    unsigned long long int offset = tstat->offset;
    // recieve chunk
cerr<<"read_thread: calling GET: offset: "<<tstat->offset<<endl;
cerr<<"read_thread: calling GET: size: "<<istat->chunk_size<<endl;
    int get_res = s.GET("",tstat->offset,istat->chunk_size,&read_callback,tstat);
    // accept last received data
    if(tstat->buffer) {
cerr<<"read_thread: calling is_read: "<<tstat->buffer_used<<endl;
      istat->buffer->is_read(tstat->buffer_h,tstat->buffer_used,tstat->offset);
    };
    tstat->offset+=tstat->buffer_used;

    if(get_res != 0) { // failed
cerr<<"read_thread: GET failed"<<endl;
      istat->buffer->error_read(true);
      break;
    } else {
      unsigned long long int cs = s.size();
      if(cs) {
        istat->content_size=cs; istat->have_content_size=true;
      };
      // if 0 bytes transfered - behind eof
cerr<<"read_thread: check for eof: "<<offset<<" - "<<tstat->offset<<endl;
      if(offset == tstat->offset) { failed=false; break; }; // eof
    };
  };
cerr<<"read_thread: loop exited"<<endl;
  istat->lock.block();
  istat->threads--;
  if(istat->threads==0) {
cerr<<"read_thread: last thread: failure: "<<failed<<endl;
    if(failed) istat->buffer->error_read(true);
    istat->buffer->eof_read(true);
  };
  tstat->s=NULL;
  istat->lock.signal_nonblock();
  istat->lock.unblock();
  return NULL;  
}

bool DataHandle::start_reading_httpg(DataBufferPar &buf) {
  if(httpg_stat == NULL) httpg_stat=new httpg_info_t;
  httpg_stat->buffer=&buf; buffer=&buf;
  httpg_stat->url=c_url.c_str();
  httpg_stat->chunk_size=buf.buffer_size();
  httpg_stat->streams=ftp_threads;
  httpg_stat->requested_size=0; // amount of data currently beeing downloaded
  httpg_stat->threads=0;
  httpg_stat->cancel=false;
  httpg_stat->have_threads=false;
  if(httpg_stat->channels) free(httpg_stat->channels);
  httpg_stat->channels=(httpg_state_t*)malloc(sizeof(httpg_state_t)*httpg_stat->streams);
  if(httpg_stat->channels==NULL) return false;
  httpg_stat->handle = this;
  httpg_stat->point = url;
  // start threads: one per stream
  httpg_stat->lock.reset();
  int started = 0;
  pthread_attr_t attr;
  if(pthread_attr_init(&attr) != 0) return false;
  if(pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED) != 0) return false;
  for(;started < httpg_stat->streams;started++) { 
    httpg_state_t* tstat = httpg_stat->channels+started;
    if(pthread_create(&(tstat->thr),&attr,&read_thread,httpg_stat) != 0) break;
  };
  pthread_attr_destroy(&attr);
  if(started==0) return false;
  return true;
}

bool DataHandle::stop_reading_httpg(void) {
  httpg_stat->lock.block();
  if(!buffer->eof_read()) {
    buffer->error_read(true);
    // cancel transfer
    httpg_stat->cancel=true;
    for(int i = 0;i<httpg_stat->streams;i++) {
      if(httpg_stat->channels[i].s) httpg_stat->channels[i].s->disconnect();
    };
  };
  while(httpg_stat->threads > 0) {
    httpg_stat->lock.wait_nonblock();
  };
  httpg_stat->lock.unblock();
  delete httpg_stat;
  return true;
}

static void* write_thread(void *arg) {
  httpg_info_t* istat = (httpg_info_t*)arg;
  istat->lock.block();
  if(istat->cancel) { istat->lock.unblock(); return NULL; };
  httpg_state_t* tstat = istat->channels+istat->threads;
  istat->threads++; istat->have_threads=true;
cerr<<"write_thread: threads: "<<istat->threads<<endl;
  HTTP_Client s(istat->url);
  tstat->s=&s;
  istat->lock.unblock();
  tstat->stat=istat;
  bool failed = true;

  if(s) for(;;) {
    /// get first buffer
    if(!istat->buffer->for_write(tstat->buffer_h,tstat->buffer_size,tstat->offset,true)) {
      if(!istat->buffer->error()) failed=false; // eof
      break;
    };
    tstat->buffer=(*(istat->buffer))[tstat->buffer_h];
    // connect
    if(s.connect() != 0) break;
    // send chunk
    int get_res = s.PUT("",tstat->offset,tstat->buffer_size,
                             (unsigned char*)(tstat->buffer),
                             istat->have_content_size?istat->content_size:0);
    if(get_res != 0) { // failed
      istat->buffer->is_notwritten(tstat->buffer_h);
      istat->buffer->error_write(true);
      break;
    } else {
      istat->buffer->is_written(tstat->buffer_h);
    };
  };
  istat->lock.block();
  istat->threads--;
  if(istat->threads==0) {
cerr<<"write_thread: last thread: failure: "<<failed<<endl;
    DataHandle* handle = istat->handle;
    DataPoint* point = istat->point;
    if(point && !failed) {
      if(!istat->buffer->checksum_valid()) { failed=true; }
      else if(point->meta()) { // pass meta information to an SE
        struct soap soap;
        HTTP_ClientSOAP s(istat->url,&soap);
        if(s.connect() != 0) {
          cerr<<"Failed to connect to "<<istat->url<<endl;
          return false;
        };
        ns__fileinfo info;
        info.size=NULL;
        // only cheksum is important here
        info.checksum=NULL;
        char checksum_[100];
        int l_ = snprintf(checksum_,99,"cksum:%16LX",istat->buffer->checksum_object()->end()); 
        checksum_[l_]=0;
        info.checksum=checksum_;
        info.acl=NULL;
        info.id=NULL; 
        info.created=NULL;
        string created("");
        if(point->meta_created_available()) {
          struct tm* t;
          struct tm tt;
          time_t t_ = point->meta_created();
          t=gmtime_r(&t_,&tt);
          if(t) {
            if(timetostring(*t,created) == 0) {
              info.created=(char*)(created.c_str());
            };
          };
        };
cerr<<"write_thread: update: created: "<<info.created<<endl;
cerr<<"write_thread: update: created: "<<created<<endl;
        ns__updateResponse rr;
        int soap_err = SOAP_OK;
        string soap_url = istat->url;
        int n = soap_url.find(':');
        // make gSOAP create http headers (should be removed in a future)
        if(n != string::npos) soap_url.replace(0,n,"http");
        if((soap_err=soap_call_ns__update(&soap,soap_url.c_str(),"update",
                                    &info,rr)) != SOAP_OK) {
          odlog(1)<<"Failed to execute remote soap call 'update' at "<<istat->url<<endl;
          failed=true;
        } else if(rr.error_code != 0) {
          odlog(1)<<"Failed ("<<rr.error_code<<") to update remote file "<<info.id<<" at "<<istat->url<<endl;
          failed=true;
        };
      };
    };
    if(failed) istat->buffer->error_write(true);
    istat->buffer->eof_write(true);
  };
  tstat->s=NULL;
  istat->lock.signal_nonblock();
  istat->lock.unblock();
  return NULL;  
}

bool DataHandle::start_writing_httpg(DataBufferPar &buf) {
  string store_url(c_url.c_str());
  // registration at SE in case of meta url ???
  if(url->meta()) { // pass meta information directly to an SE
    struct soap soap;
    HTTP_ClientSOAP s(c_url.c_str(),&soap);
    if(s.connect() != 0) {
      cerr<<"Failed to connect to "<<c_url<<endl;
      return false;
    };
    ns__fileinfo info;
    unsigned long long int size_ = url->meta_size();
    if(url->meta_size_available()) info.size=&size_;
    char checksum_[100];
    info.checksum=NULL;
    int l_ = snprintf(checksum_,99,"%16LX",url->meta_checksum()); 
    checksum_[l_]=0;
    if(url->meta_checksum_available()) info.checksum=checksum_;
    info.acl=NULL;
    string acl("");
    {
      CertInfo ci;
      acl=ci.GetDN();
      if(acl.length()) acl="<gacl><entry><person><dn>"+acl+"</dn></person><allow><read/><write/><list/><admin/></allow></entry></gacl>";
      info.acl=(char*)(acl.c_str());
    };
    info.id=strdup(url->lfn()); 
    info.created=NULL;
    if(url->meta_created_available()) {
      struct tm* t;
      struct tm tt;
      time_t t_ = url->meta_created();
      t=gmtime_r(&t_,&tt);
      if(t) {
        string created;
        if(timetostring(*t,created) == 0) {
          info.created=strdup(created.c_str());
        };
      };
    };
cerr<<"start_writing_httpg: id: "<<info.id<<endl;
cerr<<"start_writing_httpg: lfn: "<<url->lfn()<<endl;
cerr<<"start_writing_httpg: acl: "<<info.acl<<endl;
cerr<<"start_writing_httpg: created: "<<info.created<<endl;
    ns__addResponse rr;
    int soap_err = SOAP_OK;
    string soap_url = c_url;
    int n = soap_url.find(':');
    // make gSOAP create http headers (should be removed in a future)
    if(n != string::npos) soap_url.replace(0,n,"http");
    if((soap_err=soap_call_ns__add(&soap,soap_url.c_str(),"add",&info,rr))
                                                             != SOAP_OK) {
      odlog(1)<<"Failed to execute remote soap call 'add' at "<<c_url<<endl;
      if(info.id) free(info.id);
      if(info.created) free(info.created);
      return false;
    };
    if(rr.error_code != 0) {
      odlog(1)<<"Failed ("<<rr.error_code<<") to create remote file "<<info.id<<" at "<<c_url<<endl;
      if(info.id) free(info.id);
      if(info.created) free(info.created);
      return false;
    };
    if(info.id) free(info.id);
    if(info.created) free(info.created);
    if(rr.file.url) { store_url=rr.file.url; }
    else {
      if(store_url[store_url.length()-1] != '/') store_url+="/";
      if(url->lfn()[0] == '/') {
        store_url+=(url->lfn()+1);
      } else {
        store_url+=url->lfn();
      };
    };
    odlog(2)<<"Created slot at: "<<store_url<<endl;
  };
  // starting data transfer
  if(httpg_stat == NULL) httpg_stat=new httpg_info_t;
  httpg_stat->buffer=&buf; buffer=&buf;
  httpg_stat->url=strdup(store_url.c_str());
cerr<<"start_writing_httpg: url: "<<httpg_stat->url<<endl;
  httpg_stat->chunk_size=buf.buffer_size(); // not actually needed for writing
  httpg_stat->streams=ftp_threads;
cerr<<"start_writing_httpg: streams: "<<httpg_stat->streams<<endl;
  httpg_stat->requested_size=0;
  httpg_stat->threads=0;
  httpg_stat->cancel=false;
  httpg_stat->have_threads=false;
  httpg_stat->have_content_size=url->meta_size_available();
  httpg_stat->content_size=url->meta_size();
  if(httpg_stat->channels) free(httpg_stat->channels);
  httpg_stat->handle = this;
  httpg_stat->point = url;
  httpg_stat->channels=(httpg_state_t*)malloc(sizeof(httpg_state_t)*httpg_stat->streams);
cerr<<"start_writing_httpg: channels: "<<(int)httpg_stat->channels<<endl;
  if(httpg_stat->channels==NULL) return false;
  // start threads: one per stream
  httpg_stat->lock.reset();
  int started = 0;
  pthread_attr_t attr;
  if(pthread_attr_init(&attr) != 0) return false;
  if(pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED) != 0) return false;
  for(;started < httpg_stat->streams;started++) { 
cerr<<"start_reading_httpg: started: "<<started<<endl;
    httpg_state_t* tstat = httpg_stat->channels+started;
    if(pthread_create(&(tstat->thr),&attr,&write_thread,httpg_stat) != 0) break;
  };
  pthread_attr_destroy(&attr);
cerr<<"start_writing_httpg: started: "<<started<<endl;
  if(started==0) return false;
  return true;
}

bool DataHandle::stop_writing_httpg(void) {
  httpg_stat->lock.block();
  if(!buffer->eof_read()) {
    buffer->error_read(true);
    // cancel transfer
    httpg_stat->cancel=true;
    for(int i = 0;i<httpg_stat->streams;i++) {
      if(httpg_stat->channels[i].s) httpg_stat->channels[i].s->disconnect();
    };
  };
  while(httpg_stat->threads > 0) {
    httpg_stat->lock.wait_nonblock();
  };
  httpg_stat->lock.unblock();
  free((char*)(httpg_stat->url));
  delete httpg_stat;
  return true;
}

bool DataHandle::remove_httpg(void) {
cerr<<"DataHandle::remove_httpg"<<endl;
  if(!url->meta()) return false;
  struct soap soap;
  HTTP_ClientSOAP s(c_url.c_str(),&soap);
cerr<<"DataHandle::remove_httpg: created HTTP_ClientSOAP"<<endl;
  if(s.connect() != 0) {
    odlog(0)<<"Failed to connect to "<<c_url<<endl;
    return false;
  };
cerr<<"DataHandle::remove_httpg: HTTP_ClientSOAP connected"<<endl;
  ns__delResponse rr;
  int soap_err = SOAP_OK;
  string soap_url = c_url;
  soap_url+="/"; soap_url+=url->lfn();
  int n = soap_url.find(':');
  // make gSOAP create http headers (should be removed in a future)
  if(n != string::npos) soap_url.replace(0,n,"http");
cerr<<"DataHandle::remove_httpg: calling soap_call_ns__del"<<endl;
  if((soap_err=soap_call_ns__del(&soap,soap_url.c_str(),"del",rr))
                                                             != SOAP_OK) {
    odlog(1)<<"Failed to execute remote soap call 'del' at "<<c_url<<endl;
    return false;
  };
  if(rr.error_code != 0) {
    odlog(1)<<"Failed ("<<rr.error_code<<") to delete remote file "<<c_url<<endl;
    return false;
  };
cerr<<"DataHandle::remove_httpg: soap_call_ns__del finished"<<endl;
  return true;
}

static int check_callback(unsigned long long offset,unsigned long long size,char* buf,void* arg) {
  return 0;
}

bool DataHandle::check_httpg(void) {
  odlog(2)<<"check_httpg: start"<<endl;
  HTTP_Client s(c_url.c_str());
  if(s.connect() != 0) return false; // failed to connect
  int get_res = s.GET("",0,1,&check_callback,NULL); // recieve chunk
  if(get_res != 0) return false;
  return true;
}

bool DataHandle::list_files_httpg(list<DataPoint::FileInfo> &files,bool resolve) {
  odlog(0)<<"Listing through http(g,s) is not implemented yet"<<endl;
  return false;
}

