/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/

/**********************************************************************

grid_proxy_init.c

Description:
	Program used to get a session "proxy" certificate using
	your long term certificate. This is functionally equivelent 
	to the Kerberos kinit program. 

CVS Information:

	$Source: /cvs/nordugrid/grid-manager/external/voms/Socket++/newca.c,v $
	$Date: 2003/05/16 09:39:34 $
	$Revision: 1.2 $
	$Author: aleks $

**********************************************************************/

static char *rcsid = "$Header: /cvs/nordugrid/grid-manager/external/voms/Socket++/newca.c,v 1.2 2003/05/16 09:39:34 aleks Exp $";

/**********************************************************************
                             Include header files
**********************************************************************/
#if 0
#include "config.h"
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef WIN32
#include "winglue.h"
#include <io.h>
#else
#include <dirent.h>
#endif

#include "sslutils.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <sys/stat.h>

#include <openssl/buffer.h>
#include <openssl/crypto.h>
#include <openssl/objects.h>
#include <openssl/asn1.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/rsa.h>
#include <openssl/lhash.h>
#include <openssl/rand.h>

#ifdef USE_PKCS11
#include "scutils.h"
#endif

#if SSLEAY_VERSION_NUMBER >= 0x0090581fL
#include "openssl/x509v3.h"
#endif

#include "gssapi_compat.h"
#include <gssapi.h>

#include "globus_oldgaa.h"
#include "globus_oldgaa_utils.h"

#if 0
static X509 *find_issuer(X509_STORE_CTX *ctx, STACK_OF(X509) *sk, X509 *x);
static char *ca(X509 *, STACK_OF(X509) *, proxy_verify_desc *);
#endif

#define CANAMESIZE 2048

extern char caname[CANAMESIZE];

static int check_ssl_ca(const X509 *x);
static int ca_check(const X509 *x);


static void 
x509v3_cache_extensions(X509 *x)
{
	BASIC_CONSTRAINTS *bs;
	ASN1_BIT_STRING *usage;
	ASN1_BIT_STRING *ns;
	STACK_OF(ASN1_OBJECT) *extusage;
	
	int i;
	if(x->ex_flags & EXFLAG_SET) return;
#ifndef NO_SHA
	X509_digest(x, EVP_sha1(), x->sha1_hash, NULL);
#endif
	/* Does subject name match issuer ? */
	if(!X509_NAME_cmp(X509_get_subject_name(x), X509_get_issuer_name(x)))
			 x->ex_flags |= EXFLAG_SS;
	/* V1 should mean no extensions ... */
	if(!X509_get_version(x)) x->ex_flags |= EXFLAG_V1;
	/* Handle basic constraints */
	if((bs=X509_get_ext_d2i(x, NID_basic_constraints, NULL, NULL))) {
		if(bs->ca) x->ex_flags |= EXFLAG_CA;
		if(bs->pathlen) {
			if((bs->pathlen->type == V_ASN1_NEG_INTEGER)
						|| !bs->ca) {
				x->ex_flags |= EXFLAG_INVALID;
				x->ex_pathlen = 0;
			} else x->ex_pathlen = ASN1_INTEGER_get(bs->pathlen);
		} else x->ex_pathlen = -1;
		BASIC_CONSTRAINTS_free(bs);
		x->ex_flags |= EXFLAG_BCONS;
	}
	/* Handle key usage */
	if((usage=X509_get_ext_d2i(x, NID_key_usage, NULL, NULL))) {
		if(usage->length > 0) {
			x->ex_kusage = usage->data[0];
			if(usage->length > 1) 
				x->ex_kusage |= usage->data[1] << 8;
		} else x->ex_kusage = 0;
		x->ex_flags |= EXFLAG_KUSAGE;
		ASN1_BIT_STRING_free(usage);
	}
	x->ex_xkusage = 0;
	if((extusage=X509_get_ext_d2i(x, NID_ext_key_usage, NULL, NULL))) {
		x->ex_flags |= EXFLAG_XKUSAGE;
		for(i = 0; i < sk_ASN1_OBJECT_num(extusage); i++) {
			switch(OBJ_obj2nid(sk_ASN1_OBJECT_value(extusage,i))) {
				case NID_server_auth:
				x->ex_xkusage |= XKU_SSL_SERVER;
				break;

				case NID_client_auth:
				x->ex_xkusage |= XKU_SSL_CLIENT;
				break;

				case NID_email_protect:
				x->ex_xkusage |= XKU_SMIME;
				break;

				case NID_code_sign:
				x->ex_xkusage |= XKU_CODE_SIGN;
				break;

				case NID_ms_sgc:
				case NID_ns_sgc:
				x->ex_xkusage |= XKU_SGC;
			}
		}
		sk_ASN1_OBJECT_pop_free(extusage, ASN1_OBJECT_free);
	}

	if((ns=X509_get_ext_d2i(x, NID_netscape_cert_type, NULL, NULL))) {
		if(ns->length > 0) x->ex_nscert = ns->data[0];
		else x->ex_nscert = 0;
		x->ex_flags |= EXFLAG_NSCERT;
		ASN1_BIT_STRING_free(ns);
	}
	x->skid =X509_get_ext_d2i(x, NID_subject_key_identifier, NULL, NULL);
	x->akid =X509_get_ext_d2i(x, NID_authority_key_identifier, NULL, NULL);
	x->ex_flags |= EXFLAG_SET;
}

static char *
get_ca_signing_policy_path(
    const char *                        cert_dir,
    X509_NAME *                         ca_name)
{
    char *                              buffer;
    unsigned int                        buffer_len;
    unsigned long                       hash;
/*      int                                 status; */
    
    if ((cert_dir == NULL) ||
        (ca_name == NULL)) 
    {
        PRXYerr(PRXYERR_F_GET_CA_SIGN_PATH, PRXYERR_R_BAD_ARGUMENT);
        return NULL;
    }
    
    
    hash = X509_NAME_hash(ca_name);
    
    buffer_len = strlen(cert_dir) + strlen("/") + 8 /* hash */
        + strlen(SIGNING_POLICY_FILE_EXTENSION) + 1 /* NUL */;
    
    buffer = malloc(buffer_len);
    
    if (buffer == NULL) 
    {
        PRXYerr(PRXYERR_F_GET_CA_SIGN_PATH, PRXYERR_R_OUT_OF_MEMORY);
        return NULL;
    }

    sprintf(buffer,"%s%s%08lx%s", cert_dir, "/", hash,
            SIGNING_POLICY_FILE_EXTENSION);
    
    return buffer;
}

/**********************************************************************
Function:       checkstat()
Description:    check the status of a file
Parameters:
Returns:
                0 pass all the following tests
                1 does not exist
                2 not owned by user
                3 readable by someone else
                4 zero length
**********************************************************************/
static int
checkstat(
    const char*                         filename)
{
    struct stat                         stx;

    if (stat(filename,&stx) != 0)
    {
        return 1;
    }

    /*
     * use any stat output as random data, as it will 
     * have file sizes, and last use times in it. 
     */
    RAND_add((void*)&stx,sizeof(stx),2);

#if !defined(WIN32) && !defined(TARGET_ARCH_CYGWIN)
    if (stx.st_uid != getuid())
    {
#ifdef DEBUG
        fprintf(stderr,"checkstat:%s:uid:%d:%d\n",filename,
                stx.st_uid, getuid());
#endif
        return 2;
    }

    if (stx.st_mode & 066)
    {
#ifdef DEBUG
        fprintf(stderr,"checkstat:%s:mode:%o\n",filename,stx.st_mode);
#endif
        return 3;
    }
    
#endif /* !WIN32 && !TARGET_ARCH_CYGWIN */

    if (stx.st_size == 0)
    {
        return 4;
    }
    return 0;

}



#ifdef WIN32
/*********************************************************************
Function: getuid, getpid

Descriptions:
        For Windows95, WIN32, we don't have these, so we will default
    to using uid 0 and pid 0 Need to look at this better for NT. 
******************************************************************/
static unsigned long
getuid()
{
    return 0;
}

static int
getpid()
{
    return 0;
}

#endif /* WIN32 */

#if 0
int 
get_own_CA(X509_STORE_CTX *ctx, char *name, int size)
{
  X509 *x,*xtmp,*chain_ss=NULL;
  X509_NAME *xn;
  int depth,i,ok=0;
  int num;
  STACK_OF(X509) *sktmp=NULL;

  if (ctx->cert == NULL)
    return -1;

  /* first we make sure the chain we are going to build is
   * present and that the first entry is in place */
  if (ctx->chain == NULL) {
    if (((ctx->chain=sk_X509_new_null()) == NULL) ||
	(!sk_X509_push(ctx->chain,ctx->cert)))
	goto end;

    CRYPTO_add(&ctx->cert->references,1,CRYPTO_LOCK_X509);
    ctx->last_untrusted=1;
  }

  /* We use a temporary STACK so we can chop and hack at it */
  if (ctx->untrusted != NULL
      && (sktmp=sk_X509_dup(ctx->untrusted)) == NULL)

      goto end;

  num=sk_X509_num(ctx->chain);
  x=sk_X509_value(ctx->chain,num-1);
  depth=ctx->depth;

  for (;;) {
    /* If we have enough, we break */
    if (depth < num) break; /* FIXME: If this happens, we should take
			     * note of it and, if appropriate, use the
			     * X509_V_ERR_CERT_CHAIN_TOO_LONG error
			     * code later.
			     */

    /* If we are self signed, we break */
    xn=X509_get_issuer_name(x);
    if (ctx->check_issued(ctx, x,x)) break;

    /* If we were passed a cert chain, use it first */
    if (ctx->untrusted != NULL) {
      xtmp=find_issuer(ctx, sktmp,x);
      if (xtmp != NULL) {
	if (!sk_X509_push(ctx->chain,xtmp))
	  goto end;

	CRYPTO_add(&xtmp->references,1,CRYPTO_LOCK_X509);
	sk_X509_delete_ptr(sktmp,xtmp);
	ctx->last_untrusted++;
	x=xtmp;
	num++;
	/* reparse the full chain for
	 * the next one */
	continue;
      }
    }
    break;
  }

  /* at this point, chain should contain a list of untrusted
   * certificates.  We now need to add at least one trusted one,
   * if possible, otherwise we complain. */

  /* Examine last certificate in chain and see if it
   * is self signed.
   */

  i=sk_X509_num(ctx->chain);
  x=sk_X509_value(ctx->chain,i-1);
  xn = X509_get_subject_name(x);
  if (ctx->check_issued(ctx, x, x)) {
    /* we have a self signed certificate */
    if (sk_X509_num(ctx->chain) == 1) {
      /* We have a single self signed certificate: see if
       * we can find it in the store. We must have an exact
       * match to avoid possible impersonation.
       */
      ok = ctx->get_issuer(&xtmp, ctx, x);
      if ((ok <= 0) || X509_cmp(x, xtmp)) {
	ctx->error=X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT;
	ctx->current_cert=x;
	ctx->error_depth=i-1;
	if (ok == 1) X509_free(xtmp);
      }
      else {
	/* We have a match: replace certificate with store version
	 * so we get any trust settings.
	 */
	X509_free(x);
	x = xtmp;
	sk_X509_set(ctx->chain, i - 1, x);
	ctx->last_untrusted=0;
      }
    }
    else {
      /* extract and save self signed certificate for later use */
      chain_ss=sk_X509_pop(ctx->chain);
      ctx->last_untrusted--;
      num--;
      x=sk_X509_value(ctx->chain,num-1);
    }
  }

  /* We now lookup certs from the certificate store */
  for (;;) {
    /* If we have enough, we break */
    if (depth < num) break;

    /* If we are self signed, we break */
    xn=X509_get_issuer_name(x);
    if (ctx->check_issued(ctx,x,x)) break;

    ok = ctx->get_issuer(&xtmp, ctx, x);

    if (ok < 0) return ok;
    if (ok == 0) break;

    x = xtmp;
    if (!sk_X509_push(ctx->chain,x)) {
      X509_free(xtmp);
      return 0;
    }
    num++;
  }

  /* we now have our chain, lets check it... */
  xn=X509_get_issuer_name(x);

  /* Is last certificate looked up self signed? */
  if (!ctx->check_issued(ctx,x,x)) {
    if ((chain_ss == NULL) || !ctx->check_issued(ctx, x, chain_ss)) {
      if (ctx->last_untrusted >= num)
	ctx->error=X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY;
      else
	ctx->error=X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT;
      ctx->current_cert=x;
    }
    else {
      sk_X509_push(ctx->chain,chain_ss);
      num++;
      ctx->last_untrusted=num;
      ctx->current_cert=chain_ss;
      ctx->error=X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN;
      chain_ss=NULL;
    }

    ctx->error_depth=num-1;
  }
end:
  if (sktmp != NULL) sk_X509_free(sktmp);
  if (chain_ss != NULL) X509_free(chain_ss);
  if (xn) X509_NAME_oneline(xn, name, size);
  return ok;
}
#endif

/* Given a STACK_OF(X509) find the issuer of cert (if any)
 */
#if 0
static X509 *
find_issuer(X509_STORE_CTX *ctx, STACK_OF(X509) *sk, X509 *x)
{
  int i;
  X509 *issuer;
  for (i = 0; i < sk_X509_num(sk); i++) {
    issuer = sk_X509_value(sk, i);
    if (ctx->check_issued(ctx, x, issuer))
      return issuer;
  }
  return NULL;
}
#endif

/*
 * The following two functions are an ugly hack of the libglobus_gss*
 * libraries done to let a program discover the CA that issued his own
 * certificate. To do so, you must call the following function. See voms.c
 * for mode of calling.
 */
char *
getMCA(void *credential, int globusver)
{
  X509 *cert = NULL;
  X509_NAME *name;
  char *result;
  static char *buffer = NULL;
  static int   bufsize = 0;
  int ressize;
  int pos;
  int i;
  int res;
  STACK_OF(X509) *stk;

  if (!credential)
    return NULL;
#if 1
  if (globusver == 20) {
    pos = sk_X509_num(((gss_cred_id_desc *)credential)->pcd->cert_chain);
    if (pos)
      cert = sk_X509_value(((gss_cred_id_desc *)credential)->pcd->cert_chain,pos-1);
    if (!cert)
      cert = ((gss_cred_id_desc *)credential)->pcd->ucert;
  }
  else if (globusver == 22) {
    pos = sk_X509_num(((gss2_cred_id_desc *)credential)->cred_handle->cert_chain);
    if (pos)
      cert = sk_X509_value(((gss2_cred_id_desc *)credential)->cred_handle->cert_chain,pos-1);
    if (!cert)
      cert = ((gss2_cred_id_desc *)credential)->cred_handle->cert;
  }
  else
    return NULL;
#endif

#if 0
  fprintf(stderr, "globusver = %d\n", globusver);
  if (globusver == 20) {
    fprintf(stderr, "VER = 20\n");
    stk = ((gss_cred_id_desc *)credential)->pcd->cert_chain;
    pos = sk_X509_num(stk);
  }
  else if (globusver == 22) {
    fprintf(stderr, "VER = 22\n");
    stk = (((gss2_cred_id_desc *)credential)->cred_handle->cert_chain);
    pos = sk_X509_num(stk);
  }
  else
    return NULL;

  i = 0;
  while (i < pos) {
    cert = sk_X509_value(stk,i);
    x509v3_cache_extensions(cert);
    res = check_ssl_ca(cert);
    if (res == 1 || res == 3)
      break;
    i++;
  }
  if (i >= pos)
    return NULL;
#endif
  name = X509_get_issuer_name(cert);
  result = X509_NAME_oneline(name, NULL, 0);
  ressize = strlen(result);

  if (ressize >= bufsize) {
    free(buffer);
    buffer = (char *)malloc(strlen(result)+1);
      bufsize=ressize+1;
  }
  if (buffer)
    strncpy(buffer, result, ressize+1);

  OPENSSL_free(result);
  return buffer;

#if 0
  proxy_verify_desc     pvd;
  proxy_verify_ctx_desc pvxd;
  proxy_cred_desc *pcd;
  globus_gsi_cred_handle_t ggch;
  X509 *cert;
  STACK_OF(X509) *cert_chain;

  if (!credential)
    return NULL;

    /*
     * verify if the cert is good, i.e. is signed by one of the
     * trusted CAs.
     */
  proxy_verify_ctx_init(&pvxd);
  proxy_verify_init(&pvd, &pvxd);
  pvxd.certdir = "/etc/grid-security/certificates";

  if (globusver == 20) {
    pcd = ((gss_cred_id_desc *)credential)->pcd;
    cert = pcd->ucert;
    cert_chain = pcd->cert_chain;
  }
  else if (globusver == 22) {
    ggch = ((gss2_cred_id_desc *)credential)->cred_handle;
    cert = ggch->cert;
    cert_chain = ggch->cert_chain;
  }
  else
    return NULL;

  return ca(cert, cert_chain, &pvd);
#endif
}

#if 0
static char *
ca(X509 * ucert, STACK_OF(X509) *cert_chain, proxy_verify_desc *pvd)
{
    int                                 i;
    int                                 j;
    char *                              retval = NULL;
    X509_STORE *                        cert_store = NULL;
    X509_LOOKUP *                       lookup = NULL;
    X509_STORE_CTX                      csc;
    X509 *                              xcert = NULL;
    X509 *                              scert = NULL;
    static char buffer[CANAMESIZE];

    scert = ucert;
    cert_store = X509_STORE_new();
    X509_STORE_set_verify_cb_func(cert_store, proxy_verify_callback);
    if (cert_chain != NULL)
    {
        for (i=0;i<sk_X509_num(cert_chain);i++)
        {
            xcert = sk_X509_value(cert_chain,i);

            if (!scert)
            {
                scert = xcert;
            }
            else
            {
                j = X509_STORE_add_cert(cert_store, xcert);
                if (!j)
                {
                    if ((ERR_GET_REASON(ERR_peek_error()) ==
                         X509_R_CERT_ALREADY_IN_HASH_TABLE))
                    {
                        ERR_clear_error();
                        break;
                    }
                    else
                    {
                        /*DEE need errprhere */
                        goto err;
                    }
                }
            }
        }
    }
    if ((lookup = X509_STORE_add_lookup(cert_store,
                                        X509_LOOKUP_hash_dir())))
    {
        X509_LOOKUP_add_dir(lookup,pvd->pvxd->certdir,X509_FILETYPE_PEM);
        X509_STORE_CTX_init(&csc,cert_store,scert,NULL);

#if SSLEAY_VERSION_NUMBER >=  0x0090600fL
        /* override the check_issued with our version */
        csc.check_issued = proxy_check_issued;
#endif
        X509_STORE_CTX_set_ex_data(&csc,
                                   PVD_STORE_EX_DATA_IDX, (void *)pvd);
                 
        if(!get_own_CA(&csc, buffer, CANAMESIZE))
        {
            goto err;
        }
    } 
    retval = buffer;

err:
    return retval;
}
#endif

EVP_PKEY *
get_private_key(void *credential, int globusver)
{
  proxy_cred_desc *pcd;
  globus_gsi_cred_handle_t ggch;
  EVP_PKEY *pkey = NULL;

  if (!credential)
    return NULL;

  if (globusver == 20) {
    pcd = ((gss_cred_id_desc *)credential)->pcd;
    if (pcd)
      pkey = pcd->upkey;
  }
  else if (globusver == 22) {
    ggch = ((gss2_cred_id_desc *)credential)->cred_handle;
    if (ggch)
      pkey = ggch->key;
  }
  else
    return NULL;

  return pkey;
}

EVP_PKEY *
get_delegated_public_key(gss_ctx_id_t context, int globusver)
{
  EVP_PKEY *pkey = NULL;

  if (!context)
    return NULL;

  if (globusver == 20)
    pkey = X509_extract_key(((((gss_ctx_id_desc *)context)->gs_ssl)->session)->peer);
  else if (globusver == 22)
    pkey = X509_extract_key(((((gss2_ctx_id_desc *)context)->gss_ssl)->session)->peer);

  return pkey;
}

X509 *
get_delegated_cert(gss_ctx_id_t context, int globusver)
{
  if (!context)
    return NULL;

  if (globusver == 20)
    return ((((gss_ctx_id_desc *)context)->gs_ssl)->session)->peer;
  else if (globusver == 22)
    return ((((gss2_ctx_id_desc *)context)->gss_ssl)->session)->peer;
}

char *
get_peer_CA(gss_ctx_id_t context, int globusver)
{
  X509 *cert = NULL;
  X509_NAME *name;
  char *result;
  static char *buffer = NULL;
  static int   bufsize = 0;
  int ressize;
  int pos =0;

  if (!context)
    return NULL;

  if (globusver == 20)
    return caname;
  /*    cert = ((gss_ctx_id_desc *)context)->gs_ssl->session->peer; */
  else if (globusver == 22) {
    pos = sk_X509_num(((gss2_ctx_id_desc *)context)->peer_cred_handle->cred_handle->cert_chain);
    if (pos)
      cert = sk_X509_value(((gss2_ctx_id_desc *)context)->peer_cred_handle->cred_handle->cert_chain,pos-1);
    if (!cert)
      cert = ((gss2_ctx_id_desc *)context)->peer_cred_handle->cred_handle->cert;
  }
  else
    return NULL;

  name = X509_get_issuer_name(cert);
  result = X509_NAME_oneline(name, NULL, 0);
  ressize = strlen(result);

  if (ressize >= bufsize) {
    free(buffer);
    buffer = (char *)malloc(strlen(result)+1);
      bufsize=ressize+1;
  }
  if (buffer)
    strncpy(buffer, result, ressize+1);

  OPENSSL_free(result);
  return buffer;
}


#if 0
/**********************************************************************
Function: proxy_verify_callback()

Description:
        verify callback for SSL. Used to check that proxy
        certificates are only signed by the correct user, 
        and used for debuging.
        
        Also on the server side, the s3_srvr.c code does not appear
        to save the peer cert_chain, like the client side does. 
        We need these for additional proxies, so we need to 
        copy the X509 to our own stack. 

Parameters:
        ok  1 then we are given one last chance to check
                this certificate.
                0 then this certificate has failed, and ctx->error has the
                reason. We may want to override the failure. 
        ctx the X509_STORE_CTX which has as a user arg, our 
                proxy verify desc. 
   
Returns:
        1 - Passed the tests
        0 - failed.  The x509_vfy.c will return a failed to caller. 


CHANGED:
        Now memorizes the subject of the issuer CA.
**********************************************************************/

int 
proxy_verify_callback(
    int                                 ok,
    X509_STORE_CTX *                    ctx)
{
    X509_OBJECT                         obj;
    X509_CRL *                          crl;
    X509_CRL_INFO *                     crl_info;
    X509_REVOKED *                      revoked;
    STACK_OF(X509_EXTENSION) *          extensions;
    X509_EXTENSION *                    ex;
    ASN1_OBJECT *                       extension_obj;
    int                                 nid;
    char *                              s = NULL;
    SSL *                               ssl = NULL;
    proxy_verify_desc *                 pvd;
    int                                 itsaproxy = 0;
    int                                 i;
    int                                 n;
    int                                 ret;
    time_t                              goodtill;
    char *                              ca_policy_file_path = NULL;
    char *                              cert_dir            = NULL;
    int                                 ca_found = 0;
/*      char *                              ca_policy_filename  = "ca-signing-policy.conf"; */

    /*
     * If we are being called recursivly to check delegate
     * cert chains, or being called by the grid-proxy-init,
     * a pointer to a proxy_verify_desc will be 
     * pased in the store.  If we are being called by SSL,
     * by a roundabout process, the app_data of the ctx points at
     * the SSL. We have saved a pointer to the  context handle
     * in the SSL, and its magic number should be PVD_MAGIC_NUMBER 
     */
#ifdef DEBUG
    fprintf(stderr,"\nproxy_verify_callback\n");
#endif
    if ((pvd = (proxy_verify_desc *)
         X509_STORE_CTX_get_ex_data(ctx,
                                    PVD_STORE_EX_DATA_IDX)))
    {
#ifdef DEBUG
        fprintf(stderr,"Called with alternate ex_data\n");
#endif
    }
    else
    {
        ssl = (SSL *)X509_STORE_CTX_get_app_data(ctx);
        pvd = (proxy_verify_desc *)SSL_get_ex_data(ssl,
                                                   PVD_SSL_EX_DATA_IDX);
    }

    /*
     * For now we hardcode the ex_data. We could look at all 
     * ex_data to find ours. 
     * Double check that we are indeed pointing at the context
     * handle. If not, we have an internal error, SSL may have changed
     * how the callback and app_data are handled
     */

    if(pvd->magicnum != PVD_MAGIC_NUMBER)
    {
        PRXYerr(PRXYERR_F_VERIFY_CB, PRXYERR_R_BAD_MAGIC);
        return(0);
    }

#ifdef DEBUG
    fprintf(stderr," magicnum=%d,OK=%d\n",
            pvd->magicnum, ok);
    s = X509_NAME_oneline(X509_get_subject_name(ctx->current_cert),NULL,0);
    
    if (ok)
    {
        fprintf(stderr,"ctx->error_depth=%d %s\n",ctx->error_depth,
                s?s:"no-subject");
    }
    else
    {
        fprintf(stderr,"ctx->error_depth=%d error=%d %s\n",
                ctx->error_depth,ctx->error,s?s:"no-subject");
    }
    
    if (s)
    {
        free(s);
    }
    s = NULL;
    fprintf(stderr,"pvd->proxy_depth=%d, pvd->cert_depth=%d\n",
            pvd->proxy_depth, pvd->cert_depth);
#endif

    /*
     * We now check for some error conditions which
     * can be disregarded. 
     */
        
    if (!ok)
    {
        switch (ctx->error)
        {
#if SSLEAY_VERSION_NUMBER >=  0x0090581fL
        case X509_V_ERR_PATH_LENGTH_EXCEEDED:
#ifdef DEBUG
            fprintf(stderr,"X509_V_ERR_PATH_LENGTH_EXCEEDED\n");
#endif

            /* 
             * OpenSSL-0.9.5 has a bug, in that the ex_pathlen
             * may not be set if no basic constrants are present
             * but it may check for it anyway.
             */

            if (!(ctx->current_cert->ex_flags & EXFLAG_BCONS))
            {
                ok = 1;
            }
            break;
#endif
        default:
            break;
        }                       
        /* if already failed, skip the rest, but add error messages */
        if (!ok)
        {
            if (ctx->error==X509_V_ERR_CERT_NOT_YET_VALID)
            {
                PRXYerr(PRXYERR_F_VERIFY_CB,PRXYERR_R_CERT_NOT_YET_VALID);
                ERR_set_continue_needed();
            }
            else if (ctx->error==X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY)
            {
                PRXYerr(PRXYERR_F_VERIFY_CB,PRXYERR_R_LOCAL_CA_UNKNOWN); 
                ERR_set_continue_needed();
            }
            else if (ctx->error==X509_V_ERR_CERT_HAS_EXPIRED)
            {
                PRXYerr(PRXYERR_F_VERIFY_CB, PRXYERR_R_REMOTE_CRED_EXPIRED); 
                ERR_set_continue_needed();
            }

            goto fail_verify;
        }
#ifdef DEBUG
        fprintf(stderr,"proxy_verify_callback:overriding:%d\n\n",
                ctx->error);
#endif
        ctx->error = 0;
        return(ok);
    }

    /* 
     * All of the OpenSSL tests have passed and we now get to 
     * look at the certificate to verify the proxy rules, 
     * and ca-signing-policy rules. We will also do a CRL check
     */


    /*
     * Test if the name ends in CN=proxy and if the issuer
     * name matches the subject without the final proxy. 
     */
        
    ret = proxy_check_proxy_name(ctx->current_cert);
    if (ret < 0)
    {
        PRXYerr(PRXYERR_F_VERIFY_CB,PRXYERR_R_BAD_PROXY_ISSUER);
        ERR_set_continue_needed();
        ctx->error = X509_V_ERR_CERT_SIGNATURE_FAILURE;
        goto fail_verify;
    }
    if (ret > 0)
    {  /* Its a proxy */
        if (ret == 2)
        {
            /*
             * If its a limited proxy, it means it use has been limited 
             * during delegation. It can not sign other certs i.e.  
             * it must be the top cert in the chain. 
             * Depending on who we are, 
             * We may want to accept this for authentication. 
             * 
             *   Globus gatekeeper -- don't accept
             *   sslk5d accept, but should check if from local site.
             *   globus user-to-user Yes, thats the purpose 
             *    of this cert. 
             *
             * We will set the limited_proxy flag, to show we found
             * one. A Caller can then reject. 
             */

            pvd->limited_proxy = 1; /* its a limited proxy */

            if (ctx->error_depth && !pvd->multiple_limited_proxy_ok)
            {
                /* tried to sign a cert with a limited proxy */
                /* i.e. there is still another cert on the chain */
                /* indicating we are trying to sign it! */
                PRXYerr(PRXYERR_F_VERIFY_CB,PRXYERR_R_LPROXY_MISSED_USED);
                ERR_set_continue_needed();
                ctx->error = X509_V_ERR_CERT_SIGNATURE_FAILURE;
                goto fail_verify;
            }
        }

#ifdef DEBUG
        fprintf(stderr,"Passed proxy test\n");
#endif
        pvd->proxy_depth++;
        itsaproxy = 1;
    }

    if (!itsaproxy)
    {
                        
#ifdef X509_V_ERR_CERT_REVOKED
        /* 
         * SSLeay 0.9.0 handles CRLs but does not check them. 
         * We will check the crl for this cert, if there
         * is a CRL in the store. 
         * If we find the crl is not valid, we will fail, 
         * as once the sysadmin indicates that CRLs are to 
         * be checked, he best keep it upto date. 
         * 
         * When future versions of SSLeay support this better,
         * we can remove these tests. 
         * we come through this code for each certificate,
         * starting with the CA's We will check for a CRL
         * each time, but only check the signature if the
         * subject name matches, and check for revoked
         * if the issuer name matches.
         * this allows the CA to revoke its own cert as well. 
         */
        
        if (X509_STORE_get_by_subject(ctx,
                                      X509_LU_CRL, 
                                      X509_get_subject_name(ctx->current_cert),
                                      &obj))
        {
            crl =  obj.data.crl;
            crl_info = crl->crl;
#ifdef DEBUG
            {
                BIO * bp;
                bp = BIO_new_fp(stderr,BIO_NOCLOSE);
                                
                fprintf(stderr,"CRL last Update: ");
                ASN1_UTCTIME_print(bp, crl_info->lastUpdate);
                fprintf(stderr,"\nCRL next Update: ");
                ASN1_UTCTIME_print(bp, crl_info->nextUpdate);
                fprintf(stderr,"\n");
                BIO_free(bp);
            }
#endif
            /* verify the signature on this CRL */

            if (X509_CRL_verify(crl,
                                X509_get_pubkey(ctx->current_cert)) <= 0)
            {
                PRXYerr(PRXYERR_F_VERIFY_CB,PRXYERR_R_CRL_SIGNATURE_FAILURE);
                ERR_set_continue_needed();
                ctx->error = X509_V_ERR_CRL_SIGNATURE_FAILURE;
                goto fail_verify;
            }

            /* Check date see if expired */

            i = X509_cmp_current_time(crl_info->nextUpdate);
            if (i == 0)
            {
                PRXYerr(PRXYERR_F_VERIFY_CB,PRXYERR_R_CRL_NEXT_UPDATE_FIELD);
                ERR_set_continue_needed();                
                ctx->error = X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD;
                goto fail_verify;
            }
           

            if (i < 0)
            {
                PRXYerr(PRXYERR_F_VERIFY_CB,PRXYERR_R_CRL_HAS_EXPIRED);
                ERR_set_continue_needed();
                ctx->error = X509_V_ERR_CRL_HAS_EXPIRED;
                goto fail_verify;
            }
        }

        /* now check if the issuer has a CRL, and we are revoked */

#if 1
	{
	  if (!ca_found) {
	    int i = check_ssl_ca(ctx->current_cert);
	    if (i == 1 || i == 3) {
	      ca_found = 1;
	      X509_NAME_oneline(X509_get_issuer_name(ctx->current_cert), caname, CANAMESIZE);
	    }
	  }
	}
#endif

        if (X509_STORE_get_by_subject(ctx, X509_LU_CRL, 
                                      X509_get_issuer_name(ctx->current_cert),
                                      &obj))
        {
            crl = obj.data.crl;
            crl_info = crl->crl;
#ifdef DEBUG
            fprintf(stderr,"Checking  CRL\n");
#endif
            /* check if this cert is revoked */


            n = sk_X509_REVOKED_num(crl_info->revoked);
            for (i=0; i<n; i++)
            {
                revoked = (X509_REVOKED *)sk_X509_REVOKED_value(
                    crl_info->revoked,i);

                if(!ASN1_INTEGER_cmp(revoked->serialNumber,
                                     X509_get_serialNumber(ctx->current_cert)))
                {
                    long serial;
                    char buf[256];
                    PRXYerr(PRXYERR_F_VERIFY_CB,PRXYERR_R_CERT_REVOKED);
                    serial = ASN1_INTEGER_get(revoked->serialNumber);
                    sprintf(buf,"%ld (0x%lX)",serial,serial);
                    s = X509_NAME_oneline(X509_get_subject_name(
                                              ctx->current_cert),NULL,0);
                    
                    ERR_add_error_data(4,"Serial number = ",buf,
                                       " Subject=",s);

                    ctx->error = X509_V_ERR_CERT_REVOKED;
                    ERR_set_continue_needed();
#ifdef DEBUG
                    fprintf(stderr,"revolked %lX\n",
                            ASN1_INTEGER_get(revoked->serialNumber));
                                                
#endif
                    free(s);
                    s = NULL;
                    goto fail_verify;
                }
            }
        }
#endif /* X509_V_ERR_CERT_REVOKED */

        /* Do not need to check self signed certs against ca_policy_file */

        if (X509_NAME_cmp(X509_get_subject_name(ctx->current_cert),
                          X509_get_issuer_name(ctx->current_cert)))
        {
            cert_dir = pvd->pvxd->certdir ? pvd->pvxd->certdir :
                getenv(X509_CERT_DIR);

            ca_policy_file_path =
                get_ca_signing_policy_path(
                    cert_dir,
                    X509_get_issuer_name(ctx->current_cert));
                
            if (ca_policy_file_path == NULL)
            {
                PRXYerr(PRXYERR_F_VERIFY_CB, PRXYERR_R_CA_NOPATH);
                ctx->error =  X509_V_ERR_APPLICATION_VERIFICATION; 
                goto fail_verify;
            }

            /*
             * XXX - make sure policy file exists. We get a segfault later
             * if it doesn't.
             */
            if (checkstat(ca_policy_file_path) == 1)
            {
                PRXYerr(PRXYERR_F_VERIFY_CB, PRXYERR_R_CA_NOFILE);
                ctx->error = X509_V_ERR_APPLICATION_VERIFICATION;
                ERR_set_continue_needed();
                goto fail_verify;
            }

#ifdef DEBUG
            fprintf(stderr, "ca_policy_file_path is %s\n", ca_policy_file_path);
#endif /* DEBUG */

            {
                char * error_string = NULL;
                char * issuer_name;
                char * subject_name;
#ifndef NO_OLDGAA_API
                oldgaa_rights_ptr            rights          = NULL;
                oldgaa_policy_ptr            policy_handle   = NULL;
                oldgaa_answer_ptr            detailed_answer = NULL;
                oldgaa_sec_context_ptr       oldgaa_sc          = NULL;
                oldgaa_options_ptr           options         = NULL;
                oldgaa_error_code            result;
                oldgaa_data_ptr              policy_db       = OLDGAA_NO_DATA;
                uint32                       minor_status;
#else /* Von's code */
                int result;

#endif /* #ifndef NO_OLDGAA_API */


                subject_name = X509_NAME_oneline(
                    X509_get_subject_name(ctx->current_cert),
                    NULL,
                    0);
                issuer_name = X509_NAME_oneline(
                    X509_get_issuer_name(ctx->current_cert),
                    NULL,
                    0);

#ifndef NO_OLDGAA_API
 
                if(oldgaa_globus_initialize(&oldgaa_sc,
                                            &rights,
                                            &options,
                                            &policy_db,
                                            issuer_name,
                                            subject_name,
                                            ca_policy_file_path  
                       )!= OLDGAA_SUCCESS) 

                {
                    char buf[256];
                    sprintf(buf,"Minor status=%d", policy_db->error_code);
                    PRXYerr(PRXYERR_F_VERIFY_CB, PRXYERR_R_CA_POLICY_RETRIEVE);
                    ERR_add_error_data(3,buf,"\n        ",
                                       policy_db->error_str);
                    ctx->error=X509_V_ERR_APPLICATION_VERIFICATION;
                    ERR_set_continue_needed();
                    goto fail_verify;
                }


                if(oldgaa_get_object_policy_info(
                       &minor_status,  
                       OLDGAA_NO_DATA,
                       policy_db,
                       oldgaa_globus_policy_retrieve,
                       &policy_handle) != OLDGAA_SUCCESS)
                {
                    char buf[256];
                    sprintf(buf,"Minor status=%d", minor_status);
                    PRXYerr(PRXYERR_F_VERIFY_CB, PRXYERR_R_CA_POLICY_PARSE);
                    ERR_add_error_data(3,buf,"\n        ",
                                       policy_db->error_str);
                    ctx->error =  X509_V_ERR_APPLICATION_VERIFICATION;
                    ERR_set_continue_needed(); 
                    goto fail_verify;
                }

                result = oldgaa_check_authorization (&minor_status,   
                                                     oldgaa_sc,   
                                                     policy_handle,     
                                                     rights, 
                                                     options,
                                                     &detailed_answer);
                
                if (!detailed_answer)
                {
                    PRXYerr(PRXYERR_F_VERIFY_CB, PRXYERR_R_CA_POLICY_ERR);
                    ctx->error = X509_V_ERR_INVALID_PURPOSE; 
                    ERR_set_continue_needed(); 
                    if (subject_name) free(subject_name);
                    if (issuer_name) free(issuer_name);     
                    
                    oldgaa_globus_cleanup(&oldgaa_sc,
                                          &rights,
                                          options,
                                          &detailed_answer,  
                                          policy_db,
                                          NULL);
                    goto fail_verify;
                }
#ifdef DEBUG
                fprintf(stderr,
                        "oldgaa result: %d(0 yes, 1 no, -1 maybe)\n",
                        result);
                if(detailed_answer) 
                { 
                    fprintf(stderr, "\nprint detailed answer:\n\n");
#ifndef WIN32
                    if(detailed_answer->rights)
                    {
                        oldgaa_globus_print_rights(detailed_answer->rights);
                    }
#endif
                }
#endif

                if (policy_handle)
                {
                    oldgaa_release_principals(&minor_status, &policy_handle);
                }
                

                oldgaa_globus_cleanup(&oldgaa_sc,
                                      &rights,
                                      options,
                                      &detailed_answer,  
                                      policy_db,
                                      NULL);

                
#else /* Von's code */

                result = ca_policy_file_check_signature(issuer_name,
                                                        subject_name,
                                                        &error_string,
                                                        pvd->pvxd->certdir);

#endif /* #ifndef NO_OLDGAA_API */


                free(subject_name);
                free(issuer_name);

                if (result != 0)
                {
                    PRXYerr(PRXYERR_F_VERIFY_CB, PRXYERR_R_CA_POLICY_VIOLATION);

                    ctx->error = X509_V_ERR_INVALID_PURPOSE; 
                                
                    if (error_string != NULL)
                    {
                        /*
                         * Seperate error message returned from policy check
                         * from above error message with colon
                         */
                        
                        ERR_add_error_data(2, ": ", error_string);
                        free(error_string);
                    }
                    ERR_set_continue_needed();
                    goto fail_verify;
                }
                else
                {
                    if (error_string != NULL)
                    {
                        free(error_string);
                    }
                }
            }
        } /* end of do not check self signed certs */
    }

    /*
     * We want to determine the minimum amount of time
     * any certificate in the chain is good till
     * Will be used for lifetime calculations
     */

    goodtill = ASN1_UTCTIME_mktime(X509_get_notAfter(ctx->current_cert));
    if (pvd->pvxd->goodtill == 0 || goodtill < pvd->pvxd->goodtill)
    {
        pvd->pvxd->goodtill = goodtill;
    }
        
    /* We need to make up a cert_chain if we are the server. 
     * The ssl code does not save this as I would expect. 
     * This is used to create a new proxy by delegation. 
     */

    if (pvd->cert_chain == NULL)
    {
        pvd->cert_chain = sk_X509_new_null();
    }
    
    sk_X509_push(pvd->cert_chain, X509_dup(ctx->current_cert));

    pvd->cert_depth++;

    if (ca_policy_file_path != NULL)
    {
        free(ca_policy_file_path);
    }

    extensions = ctx->current_cert->cert_info->extensions;

    for (i=0;i<sk_X509_EXTENSION_num(extensions);i++)
    {
        ex = (X509_EXTENSION *) sk_X509_EXTENSION_value(extensions,i);

        if(X509_EXTENSION_get_critical(ex))
        {
            extension_obj = X509_EXTENSION_get_object(ex);

            nid = OBJ_obj2nid(extension_obj);
            
            if(nid != NID_basic_constraints &&
               nid != NID_key_usage &&
               nid != NID_ext_key_usage &&
               nid != NID_netscape_cert_type &&
               nid != NID_subject_key_identifier &&
               nid != NID_authority_key_identifier)
            {
                PRXYerr(PRXYERR_F_VERIFY_CB, PRXYERR_R_UNKNOWN_CRIT_EXT);
                ctx->error = X509_V_ERR_CERT_REJECTED;
                goto fail_verify;
            }
        }
    }

#ifdef DEBUG 
    fprintf(stderr,"proxy_verify_callback:returning:%d\n\n", ok);
#endif
    
    return(ok);

fail_verify:

    if (ctx->current_cert)
    {
        char *subject_s = NULL;
        char *issuer_s = NULL;
                
        subject_s = X509_NAME_oneline(
            X509_get_subject_name(ctx->current_cert),NULL,0);
        issuer_s = X509_NAME_oneline(
            X509_get_issuer_name(ctx->current_cert),NULL,0);
        
        switch (ctx->error)
        {
            case X509_V_OK:
            case X509_V_ERR_INVALID_PURPOSE:
            case X509_V_ERR_APPLICATION_VERIFICATION:
                 PRXYerr(PRXYERR_F_VERIFY_CB,PRXYERR_R_CB_ERROR_MSG);
                 ERR_add_error_data(6, 
                    "\n        File=", 
                    ca_policy_file_path ? ca_policy_file_path : "UNKNOWN",
                    "\n        subject=",
                    subject_s ? subject_s : "UNKNOWN",
                    "\n        issuer =",
                    issuer_s ? issuer_s : "UNKNOWN");
            break;
            case X509_V_ERR_CERT_NOT_YET_VALID:
            case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
            case X509_V_ERR_CERT_HAS_EXPIRED:
                 PRXYerr(PRXYERR_F_VERIFY_CB,PRXYERR_R_CB_ERROR_MSG);
                 ERR_add_error_data(4, 
                    "\n        subject=",
                    subject_s ? subject_s : "UNKNOWN",
                    "\n        issuer =",
                    issuer_s ? issuer_s : "UNKNOWN");
            break;
            case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
                 PRXYerr(PRXYERR_F_VERIFY_CB,PRXYERR_R_CA_UNKNOWN);
                    ERR_add_error_data(2, "\n        issuer =",
                    issuer_s ? issuer_s : "UNKNOWN");
            break;

            default:
                PRXYerr(PRXYERR_F_VERIFY_CB,PRXYERR_R_CB_CALLED_WITH_ERROR);
                ERR_add_error_data(6,"\n        error =",
                    X509_verify_cert_error_string(ctx->error),
                    "\n        subject=",
                    subject_s ? subject_s : "UNKNOWN",
                    "\n        issuer =",
                    issuer_s ? issuer_s : "UNKNOWN");
        }

        free(subject_s);
        free(issuer_s);
    }
    if (ca_policy_file_path != NULL)
    {
        free(ca_policy_file_path);
    }

    return(0);

}
#endif

#define V1_ROOT (EXFLAG_V1|EXFLAG_SS)
#define ku_reject(x, usage) \
	(((x)->ex_flags & EXFLAG_KUSAGE) && !((x)->ex_kusage & (usage)))
#define xku_reject(x, usage) \
	(((x)->ex_flags & EXFLAG_XKUSAGE) && !((x)->ex_xkusage & (usage)))
#define ns_reject(x, usage) \
	(((x)->ex_flags & EXFLAG_NSCERT) && !((x)->ex_nscert & (usage)))

static int ca_check(const X509 *x)
{
	/* keyUsage if present should allow cert signing */
	if(ku_reject(x, KU_KEY_CERT_SIGN)) return 0;
	if(x->ex_flags & EXFLAG_BCONS) {
		if(x->ex_flags & EXFLAG_CA) return 1;
		/* If basicConstraints says not a CA then say so */
		else return 0;
	} else {
		if((x->ex_flags & V1_ROOT) == V1_ROOT) return 3;
		/* If key usage present it must have certSign so tolerate it */
		else if (x->ex_flags & EXFLAG_KUSAGE) return 3;
		else return 2;
	}
}

/* Check SSL CA: common checks for SSL client and server */
static int check_ssl_ca(const X509 *x)
{
	int ca_ret;
	ca_ret = ca_check(x);
	if(!ca_ret) return 0;
	/* check nsCertType if present */
	if(x->ex_flags & EXFLAG_NSCERT) {
		if(x->ex_nscert & NS_SSL_CA) return ca_ret;
		return 0;
	}
	if(ca_ret != 2) return ca_ret;
	else return 0;
}

IMPLEMENT_STACK_OF(X509)
IMPLEMENT_ASN1_SET_OF(X509)

IMPLEMENT_STACK_OF(X509_NAME)

IMPLEMENT_STACK_OF(X509_ATTRIBUTE)
IMPLEMENT_ASN1_SET_OF(X509_ATTRIBUTE)
