#include "../../std.h"

#include "util.h"

#include "util++.h"

GACLperm GACLtestFileAclForVOMS(const char *filename,const char* subject,const vector<struct voms> &voms_data) {
  const char* name;
  int path_l;
  char* gname;
  GACLacl* acl = NULL;
  struct stat st;
  GACLuser* user = NULL;
  GACLperm perm;
  GACLcred* cred = NULL;
  const char* gacl_name = ".gacl-";

  if(subject[0] == 0) return GACL_PERM_NONE;
  gname=GACLmakeName(filename);
  if(gname == NULL) return GACL_PERM_NONE;
  if(stat(gname,&st) == 0) {
    if(!S_ISREG(st.st_mode)) { free(gname); return GACL_PERM_NONE; };
    acl=GACLloadAcl(gname);
  } else {
    acl=GACLloadAclForFile((char*)filename);
  };
  free(gname);
  if(acl == NULL) return GACL_PERM_NONE;
  cred=GACLnewCred("person");
  if(cred==NULL) { GACLfreeAcl(acl); return GACL_PERM_NONE; };
  if(!GACLaddToCred(cred,"dn",(char*)subject)) goto err_exit;
  /* GACLnewUser does not create copy of cred. But GACLfreeUser destoyrs cred */  user=GACLnewUser(cred);
  if(user == NULL) goto err_exit;
  for(vector<struct voms>::const_iterator v = voms_data.begin();v!=voms_data.end();++v) {
    for(vector<struct data>::const_iterator u = v->std.begin();u!=v->std.end();++u) {
      cred=GACLnewCred("voms-cred");
      if(cred==NULL) goto err_exit;
      if(!GACLaddToCred(cred,"voms",(char*)(v->server.c_str()))) goto err_exit;
      if(!GACLaddToCred(cred,"vo",(char*)(v->voname.c_str()))) goto err_exit;
      if(!GACLaddToCred(cred,"group",(char*)(u->group.c_str()))) goto err_exit;
      if(!GACLaddToCred(cred,"role",(char*)(u->role.c_str()))) goto err_exit;
      if(!GACLaddToCred(cred,"capability",(char*)(u->cap.c_str()))) goto err_exit;
      if(!GACLuserAddCred(user,cred)) goto err_exit;
      cred=NULL;
    };
  };
  perm=GACLtestUserAcl(acl,user);
  GACLfreeAcl(acl);
  GACLfreeUser(user);
  return perm;
err_exit:
  if(acl) GACLfreeAcl(acl);
  if(cred) GACLfreeCred(cred);
  if(user) GACLfreeUser(user);
  return GACL_PERM_NONE;
}
