#include "../../std.h"
#include <globus_common.h>
#include <globus_io.h>
#include <fstream>

#include "../../files/info_types.h"
#include "../../misc/log_time.h"
#include "../../misc/globus_error_utils.h"

#include "nl.nsmap"

globus_io_handle_t s;
globus_io_attr_t attr;
bool test = false;


int logger(const char* url,const char* cluster,char const * const * dirs,int num);

int local_fsend(struct soap *sp, const char* buf, size_t l) {
  if(sp->socket != 0) return SOAP_SSL_ERROR;
  size_t n = 0;
  for(;n<l;) {
    globus_size_t ll = 0;
    globus_result_t res;
    if((res=globus_io_write(&s,(globus_byte_t*)buf+n,l-n,&ll))!=GLOBUS_SUCCESS){
      odlog(0)<<"Failed to write to network: "<<res<<endl;
      return SOAP_SSL_ERROR;
    };
    n+=ll;
  };
  return SOAP_OK;
}

size_t local_frecv(struct soap* sp, char* buf, size_t l) {
  if(sp->socket != 0) return SOAP_SSL_ERROR;
  globus_size_t ll = 0;
  globus_result_t res;
  if((res=globus_io_read(&s,(globus_byte_t*)buf,l,1,&ll)) != GLOBUS_SUCCESS) {
    odlog(0)<<"Failed to read from network: "<<res<<endl;
    return 0;
  };
  return ll;
}

int local_fopen(struct soap* sp, const char* endpoint, const char* host, int port) {
  globus_result_t res;
  if(sp->socket == 0) {
    odlog(1)<<"Warning: non-expected multiple soap_fopen detected"<<endl;
    globus_io_close(&s);
    sp->socket=-1;
  };
  if((res=globus_io_tcp_connect((char*)host,port,&attr,&s)) != GLOBUS_SUCCESS) {
    odlog(0)<<"Failed to connect: "<<res<<endl;
    return -1;
  };
  odlog(1)<<"Connected to remote site"<<endl;
  sp->socket=0;
  return 0;
};

int local_fclose(struct soap* sp) {
  if(sp->socket == 0) {
    globus_io_close(&s);
    sp->socket=-1;
  };
  return SOAP_OK;
};

#ifdef OLD_GLOBUS_IO
globus_bool_t auth_cb(void* arg,globus_io_handle_t* h,globus_result_t result,char* identity,gss_ctx_id_t* context_handle) {
#else
globus_bool_t auth_cb(void* arg,globus_io_handle_t* h,globus_result_t result,char* identity,gss_ctx_id_t context_handle) {
#endif
  odlog(1)<<"Remote identity: "<<identity<<endl;
  return GLOBUS_TRUE;
}

int main(int argc,char* argv[]) {
  char* url = "http://localhost:8000/logger";
  char* cluster = "localhost";

  opterr=0;
  int n;
  while((n=getopt(argc,argv,"htc:l:u:")) != -1) {
    switch(n) {
      case ':': { cerr<<"Missing argument\n"; return 1; };
      case '?': { cerr<<"Unrecognized option\n"; return 1; };
      case 't': {
        test=true; 
      }; break;
      case 'h': {
        cout<<"logger [-c cluster_name] [-u url] control_dir ..."<<endl;
         return 0;
      };
      case 'c': {
        cluster=optarg;
      }; break;
      case 'u': {
        url=optarg;
      }; break;
      default: { cerr<<"Options processing error\n"; return 1; };
    };
  };
  string soap_url = url;
  // preparing url for gSOAP
  if(strncasecmp(soap_url.c_str(),"http://",7) == 0) {
  } else if(strncasecmp(soap_url.c_str(),"https://",8) == 0) {
    soap_url.replace(0,5,"http");
  } else if(strncasecmp(soap_url.c_str(),"httpg://",8) == 0) {
    soap_url.replace(0,5,"http");
  } else {
    odlog(0)<<"Wrong url: "<<soap_url<<endl; return 1;
  };
  return logger(soap_url.c_str(),cluster,argv+optind,argc-optind);
}

int logger(const char* url,const char* cluster,char const * const * dirs,int num) {
  int ret = -1;

  int r;
  struct soap soap;
  /* initialize globus io connection related objects */
  globus_io_secure_authorization_data_t auth;
  if(globus_module_activate(GLOBUS_IO_MODULE) != GLOBUS_SUCCESS) exit(1);
  globus_io_tcpattr_init(&attr);
  globus_io_attr_set_secure_authentication_mode(&attr,GLOBUS_IO_SECURE_AUTHENTICATION_MODE_GSSAPI,GSS_C_NO_CREDENTIAL);
  globus_io_secure_authorization_data_initialize(&auth);
  globus_io_secure_authorization_data_set_callback(&auth,&auth_cb,NULL);
  globus_io_attr_set_secure_authorization_mode(&attr,GLOBUS_IO_SECURE_AUTHORIZATION_MODE_HOST,GLOBUS_NULL);
  globus_io_attr_set_secure_channel_mode(&attr,GLOBUS_IO_SECURE_CHANNEL_MODE_GSI_WRAP);
  globus_io_attr_set_secure_protection_mode(&attr,GLOBUS_IO_SECURE_PROTECTION_MODE_PRIVATE);
  // globus_io_attr_set_secure_protection_mode(&attr,GLOBUS_IO_SECURE_PROTECTION_MODE_SAFE);
  globus_io_attr_set_secure_delegation_mode(&attr,GLOBUS_IO_SECURE_DELEGATION_MODE_FULL_PROXY);
  globus_io_attr_set_secure_proxy_mode(&attr,GLOBUS_IO_SECURE_PROXY_MODE_NONE);
  globus_result_t res;

  /* intialize soap */
  soap_init(&soap);
  soap.fsend=&local_fsend;
  soap.frecv=&local_frecv;
  soap.fopen=&local_fopen;
  soap.fclose=&local_fclose;
  soap.http_version="1.1";
  soap.keep_alive=1;
  soap_set_imode(&soap,SOAP_IO_KEEPALIVE);
  soap_set_omode(&soap,SOAP_IO_KEEPALIVE);
//  soap.keep_alive=0;

  int soap_err = SOAP_OK;
  if(test) {
    if((res=globus_io_tcp_connect((char*)"localhost",8000,&attr,&s))!=GLOBUS_SUCCESS){
      odlog(0)<<"Failed to connect to localhost: "<<res<<endl;
      return -1;
    };
    globus_io_close(&s);
  } else for(int n = 0;dirs[n] && n<num;n++) {
    string logdir(dirs[n]); logdir+="/logs/";
    struct dirent file_;
    struct dirent *file;
    odlog(1)<<"Processing directory: "<<logdir<<endl;
    DIR *dir=opendir(logdir.c_str());
    if(dir == NULL) continue;
    for(;;) {
      readdir_r(dir,&file_,&file);
      if(file == NULL) break;
      string logfile = logdir+file->d_name;
      struct stat st;
      if(stat(logfile.c_str(),&st) != 0) continue;
      if(!S_ISREG(st.st_mode)) continue;
      odlog(1)<<"Processing file: "<<logfile<<endl;
      /* read file */
      char buf[4096];
      ifstream f(logfile.c_str());
      if(!f.is_open() ) continue; /* can't open file */
      nl__jobinfo info;  // TODO: constructor/destructor
      info.start=0; info.end=0; info.cluster=NULL; info.user=NULL;
      info.id=NULL; info.name=NULL; info.failure=NULL; info.lrms=NULL;
      info.queue=NULL; info.rsl=NULL; info.ui=NULL;
      info.usedcpu=0; info.usedmem=0;
      for(;!f.eof();) {
        istream_readline(f,buf,sizeof(buf));
        char* value = strchr(buf,'=');
        if(!value) continue;
        char* name=buf; (*value)=0; value++;
        #define COMPARE_AND_SET(NM,ATTR) \
          if(!strcmp(name,NM)) { info.ATTR=strdup(value); continue; }
        info.cluster=strdup(cluster);
        odlog(2)<<"name: "<<name<<" - value: "<<value<<endl;
        if(!strcmp(name,"starttime")) {
          mds_time t; t=value; if(t != (time_t)(-1)) info.start=t;
          continue; 
        };
        if(!strcmp(name,"endtime")) {
          mds_time t; t=value; if(t != (time_t)(-1)) info.end=t;
          continue; 
        };
        COMPARE_AND_SET("gridid",id);
        COMPARE_AND_SET("lrms",lrms);
        COMPARE_AND_SET("queue",queue);
        COMPARE_AND_SET("subject",user);
        COMPARE_AND_SET("jobname",name);
        COMPARE_AND_SET("failure",failure);
        COMPARE_AND_SET("rsl",rsl);
        COMPARE_AND_SET("clientname",ui);
      };
      f.close();
      int r = 1;
      if((soap_err=soap_call_nl__add(&soap,url,"add",&info,r)) != SOAP_OK) {
        odlog(1)<<"Failed to pass information to database"<<endl; r=1;
      } else if(r != 0) {
        odlog(1)<<"Record refused. Error code: "<<r<<endl;
      };
      soap_destroy(&soap);
      soap_end(&soap);
      if(info.cluster) free(info.cluster);
      if(info.user)free(info.user);
      if(info.id) free(info.id);
      if(info.name) free(info.name);
      if(info.failure) free(info.failure);
      if(info.lrms) free(info.lrms);
      if(info.queue) free(info.queue);
      if(info.rsl) free(info.rsl);
      if(info.ui) free(info.ui);
      if(soap_err != SOAP_OK) break;
      if(r == 0) {
        unlink(logfile.c_str());
      };
    };
    closedir(dir);
    if(soap_err != SOAP_OK) break;
  };
  soap_done(&soap);
  local_fclose(&soap); // just in case
  globus_io_secure_authorization_data_destroy(&auth);
  globus_io_tcpattr_destroy(&attr);
  globus_module_deactivate(GLOBUS_IO_MODULE);
  return ret;
}

