#include "../../std.h"

#ifndef HAVE_OLD_LIBXML_INCLUDES
#include <libxml/parser.h>
#else
#include <parser.h>
#ifndef xmlChildrenNode
#define xmlChildrenNode childs
#endif
#endif

// #include "gacl.h"
#include "file.h"


int SEFile::check_acl(const AuthUser &user) {
  string gname=path+".acl";
  if(user.DN()[0] == 0) return FILE_ACC_NONE;
  GACLacl* acl;
  struct stat st;
  if(stat(gname.c_str(),&st) == 0) {
    if(!S_ISREG(st.st_mode)) return FILE_ACC_NONE;
    acl=GACLloadAcl((char*)(gname.c_str()));
  } else {
    return FILE_ACC_ALL;
  };
  if(acl == NULL) return FILE_ACC_NONE;
  GACLcred* cred = GACLnewCred("person");
  if(cred==NULL) { GACLfreeAcl(acl); return FILE_ACC_NONE; };
  if(!GACLaddToCred(cred,"dn",(char*)(user.DN()))) {
    GACLfreeCred(cred); GACLfreeAcl(acl); return FILE_ACC_NONE;
  };
  /* GACLnewUser does not create copy of cred. But GACLfreeUser destoyrs cred */
  GACLuser* guser = GACLnewUser(cred);
  if(guser == NULL) {
    GACLfreeAcl(acl); GACLfreeCred(cred); return FILE_ACC_NONE;
  };
  GACLperm perm = GACLtestUserAcl(acl,guser);
  GACLfreeAcl(acl);
  GACLfreeUser(guser);
  return perm;
}

extern "C" {
extern GACLentry *GACLparseEntry(xmlNodePtr cur);
};

static GACLacl *GACLacquireAcl(char *str)
{
  xmlDocPtr   doc;
  xmlNodePtr  cur;
  GACLacl    *acl;
  GACLentry  *entry;

  doc = xmlParseMemory(str,strlen(str));
  if (doc == NULL) return NULL;

  cur = xmlDocGetRootElement(doc);

  if (xmlStrcmp(cur->name, (const xmlChar *) "gacl"))
    {
      free(doc);
      free(cur);
      return NULL;
    }

  cur = cur->xmlChildrenNode;

  acl = GACLnewAcl();

  while (cur != NULL)
       {
         entry = GACLparseEntry(cur);
         if (entry == NULL)
           {
             GACLfreeAcl(acl);
             xmlFreeDoc(doc);
             return NULL;
           }

         GACLaddEntry(acl, entry);

         cur=cur->next;
       }

  xmlFreeDoc(doc);
  return acl;
}

int SEFile::write_acl(const AuthUser &user,const char* acl) {
//  if(!(check_acl(user) & GACL_PERM_ADMIN)) return -1;
  if((!acl) || (!*acl)) return -1;
  GACLacl* acl_ = GACLacquireAcl((char*)acl);
  if(acl_ == NULL) return -1;
  string gname=path+".acl";
  if(!GACLsaveAcl((char*)(gname.c_str()),acl_)) { GACLfreeAcl(acl_); return -1; };
  return 0;
}

int SEFile::read_acl(const AuthUser &user,string& acl) {
  string gname=path+".acl";
  int h=::open(gname.c_str(),O_RDONLY);
  if(h==-1) return -1;
  int l;
  for(;;) {
    char acl_buf[1024];
    int ll = ::read(h,acl_buf,sizeof(acl_buf)-1);
    if(ll == -1) { close(h); return -1; };
    if(l>=10240) { close(h); return -1; };
    if(ll == 0) break;
    acl_buf[ll]=0; acl.append(acl_buf);
    l+=ll;
  };
  close(h);
  return 0;
}

