#include "std.h"
#include "datamove/datapoint.h"
#include "datamove/datamove.h"
#include "misc/log_time.h"

#ifdef NG_WRAPPER
extern "C" {
#endif
int ngacl(int argc,char** argv) {
  LogTime::active=false;
  bool verbose = false;
  for(;;) {
    opterr=0;
    char optc=getopt(argc,argv,"+hvd:c");
    if(optc == -1) break;
    switch(optc) {
      case 'h': {
        olog<<"ngacl [-h] [-v] [-d level] get|set url"<<endl;
        olog<<" -h - print help"<<endl;
        olog<<" -v - print version"<<endl;
        olog<<" -d - set debug level"<<endl;
        exit(1);
      }; break;
      case 'v': {
        olog<<"ngacl: version "<<VERSION<<endl;
        exit(0);
      };
      case 'd': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<0)) {
          olog<<"Improper debug level '"<<optarg<<"'"<<endl;
          exit(1);
        };
        verbose=true;
        LogTime::level=i;
      }; break;
      case '?': {
        olog<<"Unsupported option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      case ':': {
        olog<<"Missing parameter for option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      default: {
        olog<<"Undefined processing error"<<endl;
        exit(1);
      };
    };
  };
  char* command = argv[optind+0];
  if(command == NULL) {
    olog<<"Missing action" << endl; return 1;
  };
  char* file_url = argv[optind+1];
  if(file_url == NULL) {
    olog<<"Missing url" << endl; return 1;
  };
  if(argv[optind+2] != NULL) {
    olog<<"Unrecognized option: "<<argv[optind+1]<<endl; return 1;
  };
  if(strncmp(file_url,"gsiftp://",9)) {
    olog<<"Only GridFTP servers with GACL are supported"<<endl; return 1;
  };
  string gacl_url(file_url);
  int n = gacl_url.rfind('/');
  if(n < 9) {
    olog<<"URL contains no path"<<endl; return 1;
  };
  if(n == (gacl_url.length()-1)) {
    gacl_url.insert(n+1,".gacl");
  } else {
    gacl_url.insert(n+1,".gacl-");
  };
  DataPoint url(gacl_url.c_str());
  char* tmp_name = strdup("/tmp/ngacl.XXXXXX");
  int tmp_h = mkstemp(tmp_name);
  if(tmp_h == -1) {
    olog<<"Failed to create temporary file"<<endl; return 1;
  };
  fchmod(tmp_h,S_IRUSR | S_IWUSR);
  close(tmp_h);
  char* tmp_url = (char*)malloc(7+strlen(tmp_name)+1);
  if(tmp_url == NULL) {
    unlink(tmp_name); free(tmp_name);
    olog<<"Failed to allocate memory"<<endl; return 1;
  };
  strcpy(tmp_url,"file://"); strcat(tmp_url,tmp_name);
  DataPoint tmp(tmp_url); free(tmp_url);
  DataMove mover;
  mover.secure(false);
  mover.passive(true);
  mover.verbose(verbose);
  DataCache cache;
  if(!strcmp(command,"get")) {
    if(mover.Transfer(url,tmp,cache,UrlMap()) != DataMove::success) {
      olog<<"ACL transfer FAILED."<<endl;
      unlink(tmp_name); free(tmp_name);
      return 1;
    };
    tmp_h=::open(tmp_name,O_RDONLY);
    if(tmp_h == -1) {
      unlink(tmp_name); free(tmp_name); 
      olog<<"Failed to open temporary file"<<endl; return 1;
    };
    char buf[256];
    for(;;) {
      int l = ::read(tmp_h,buf,sizeof(buf)-1);
      if(l==0) break;
      if(l==-1) {
        close(tmp_h); unlink(tmp_name); free(tmp_name); 
        olog<<"Failed to read temporary file"<<endl; return 1;
      };
      buf[l]=0;
      cout<<buf;
    };
    close(tmp_h);
  } else if(!strcmp(command,"set")) {
    tmp_h=::open(tmp_name,O_WRONLY);
    if(tmp_h == -1) {
      unlink(tmp_name); free(tmp_name);
      olog<<"Failed to open temporary file"<<endl; return 1;
    };
    char buf[256];
    for(;;) {
      int l = ::read(0,buf,sizeof(buf));
      if(l==0) break;
      if(l==-1) {
        close(tmp_h); unlink(tmp_name); free(tmp_name);
        olog<<"Failed to read stdin"<<endl; return 1;
      };
      char* p = buf;
      for(;l>0;) {
        int ll=write(tmp_h,p,l);
        if(ll==-1) {
          close(tmp_h); unlink(tmp_name); free(tmp_name);
          olog<<"Failed to write temporary file"<<endl; return 1;
        };
        l-=ll;
      };
    };
    close(tmp_h);
    if(mover.Transfer(tmp,url,cache,UrlMap()) != DataMove::success) {
      olog<<"ACL transfer FAILED."<<endl;
      unlink(tmp_name); free(tmp_name);
      return 1;
    };
  } else {
    unlink(tmp_name); free(tmp_name);
    olog<<"Only get and set commands are supported"<<endl; return 1;
  };
  unlink(tmp_name);
  if(verbose) olog<<"Operation succeeded"<<endl;
  return 0;
}
#ifdef NG_WRAPPER
};
#endif

