#include "std.h"
#include "datamove/datacache.h"
#include "datamove/datamove.h"
#include "misc/globus_modules.h"
#include "misc/log_time.h"


#ifdef NG_WRAPPER
extern "C" {
#endif
int ngcopy(int argc,char** argv) {
  LogTime::active=false;
  LogTime::level=0;
  bool verbose = false;
  char* cache_path = NULL;
  char* cache_data_path = NULL;
  char* id = "<ngcopy>";
  bool secure = false;
  bool passive = false;
  bool force = false;
  for(;;) {
    opterr=0;
    char optc=getopt(argc,argv,"+hvpfd:sc:C:");
    if(optc == -1) break;
    switch(optc) {
      case 'h': {
        olog<<"ngcopy [-h] [-v] [-d level] [-s] [-c cache [-C data]] source_url destination_url"<<endl;
        olog<<" -h - print help"<<endl;
        olog<<" -v - print version"<<endl;
        olog<<" -d - set debug level (use 0 for verbose)"<<endl;
        olog<<" -c - path to local cache (use to put file into cache)"<<endl;
        olog<<" -C - path for cache data (if different from -c)"<<endl;
        olog<<" -s - use secure transfer (insecure by default)"<<endl;
        olog<<" -p - use passive transfer (does not work if secure is on)"<<endl;
        olog<<" -f - force transfer to metaurl even if LFN exists"<<endl;
        exit(1);
      }; break;
      case 'v': {
        olog<<"ngcopy: version "<<VERSION<<endl;
        exit(0);
      };
      case 'd': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<0)) {
          olog<<"Improper debug level '"<<optarg<<"'"<<endl;
          exit(1);
        };
        verbose=true;
        LogTime::level=i;
      }; break;
      case 's': {
        secure=true;
      }; break;
      case 'p': {
        passive=true;
      }; break;
      case 'f': {
        force=true;
      }; break;
      case 'c': {
        cache_path=optarg;
      }; break;
      case 'C': {
        cache_data_path=optarg;
      }; break;
      case '?': {
        olog<<"Unsupported option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      case ':': {
        olog<<"Missing parameter for option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      default: {
        olog<<"Undefined processing error"<<endl;
        exit(1);
      };
    };
  };
  char* source_url = argv[optind+0];
  if(source_url == NULL) {
    olog<<"Missing source url" << endl; return 1;
  };
  char* destination_url = argv[optind+1];
  if(destination_url == NULL) {
    olog<<"Missing destination url" << endl; return 1;
  };
  if(argv[optind+2] != NULL) {
    olog<<"Unrecognised options: "<<argv[optind+2]<<endl; return 1;
  };
  DataPoint source(source_url);
  DataPoint destination(destination_url);
  DataMove mover;
  mover.secure(secure);
  mover.passive(passive);
  mover.verbose(verbose);
  mover.force_to_meta(force);
  DataCache cache(cache_path,cache_data_path,NULL,id,getuid(),getgid());
  if(mover.Transfer(source,destination,cache,UrlMap()) != DataMove::success) {
    cerr<<"Transfer FAILED."<<endl;
    return 1;
  }
  else {
    if(verbose) cout<<"Transfer complete."<<endl;
    return 0;
  };
}
#ifdef NG_WRAPPER
};
#endif

