#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <string>
#include <vector>
#include <map>

#include "LdapQuery.h"
#include "Environment.h"

#ifndef UI_MDSQUERY
#define UI_MDSQUERY

#ifndef UNDEFINED
#define UNDEFINED -1
#endif

#if __CYGWIN__ || hpux
#undef stdin
#undef stdout
#undef stderr
#endif


class Giis;


class Job {

 public:
  Job (const string & id_);
  ~Job () {};
  const string & GetId () const;
  const string & GetStatus () const;
  const string & GetName () const;
  const string & GetStdout () const;
  const string & GetStderr () const;
  const string & GetGmlog () const;
  void SetAttr (const string & attr, const string & value);
  void Print (ostream & s = cout) const;
  void PrintShort (ostream & s = cout) const;
  bool IsStatus (const vector <string> & stat);
 private:
  string id;
  string owner;
  string execcluster;
  string execqueue;
  string stdout;
  string stderr;
  string stdin;
  string gmlog;
  int reqcput;
  string status;
  int queuerank;
  string lrmscomment;
  string submissionui;
  string submissiontime;
  int usedcputime;
  int usedwalltime;
  string erasetime;
  int usedmem;
  string errors;
  string jobname;
  vector <Environment> runtimeenv;
  int cpucount;
  vector <string> execnodes;
};


class Queue {

 public:
  Queue (const string & name_);
  ~Queue () {};
  const string & GetName () const;
  const string & GetStatus () const;
  int GetRunning () const;
  int GetQueued () const;
  int GetMaxRunning () const;
  int GetMaxQueuable () const;
  int GetMaxCpuTime () const;
  int GetMinCpuTime () const;
  int GetDefaultCpuTime () const;
  int GetTotalCpus () const;
  int GetNodeMemory () const;
  const string & GetArchitecture () const;
  int GetFrequency () const;
  int GetUserFreeCpus (int minutes = 0) const;
  long long int GetUserDiskSpace () const;
  void ClaimUserDiskSpace (long long int);
  const Environment * FindEnvironment (const string & attr,
				       const EnvironmentTest & envtest) const;
  void SetAttr (const string & attr, const string & value);
  void Print (ostream & s = cout) const;
  void PrintShort (ostream & s = cout) const;
  void Accept (int minutes, int count);
 private:
  string name;
  string status;
  int running;
  int queued;
  int maxrunning;
  int maxqueuable;
  int maxuserrun;
  int maxcputime;
  int mincputime;
  int defaultcputime;
  string schedpolicy;
  int totalcpus;
  string nodecpu;
  int nodemem;
  string arch;
  vector <Environment> opsys;
  int gridrunning;
  int gridqueued;
  vector <Environment> runtimeenv;
  vector <Environment> middleware;
  map <int, int> userfreecpus;
  long long int userdiskspace;
  int userqueuelength;
 public:
  vector <Job> jobs;
};


class Cluster {

 public:
  Cluster (const string & name_);
  ~Cluster () {};
  const string & GetName () const;
  const string   GetLongName () const;
  const string & GetContact () const;
  const string & GetArchitecture () const;
  int GetFrequency () const;
  int GetNodeMemory () const;
  int GetTotalCpus () const;
  int GetQueued () const;
  long long int GetCacheFree () const;
  void ClaimCache (long long int size);
  long long int GetSessionDirFree () const;
  void ClaimSessionDir (long long int size);
  const Environment * FindEnvironment (const string & attr,
				       const EnvironmentTest & envtest) const;
  bool MatchLocalSe (const string & location) const;
  bool HaveNodeAccess (const string & attr) const;
  void SetAttr (const string & attr, const string & value);
  void Print (ostream & s = cout) const;
  void PrintShort (ostream & s = cout) const;
  int Connect (const string & usersn, bool anonymous,
	       int timeout, int debug = 0);
  int Query (const string & usersn, int timeout, int debug = 0);
  int Result (int timeout, int debug = 0);
  int Find (const string & usersn, bool anonymous, int timeout, int debug = 0);
 private:
  string name;
  string alias;
  string contact;
  vector <string> support;
  string lrmstype;
  string lrmsversion;
  string lrmsconfig;
  string arch;
  vector <Environment> opsys;
  bool homogeneity;
  string nodecpu;
  int nodemem;
  int totalcpus;
  map <int, int> cpudist;
  long long int sessiondirfree;
  long long int sessiondirtotal;
  long long int cachefree;
  long long int cachetotal;
  vector <Environment> runtimeenv;
  vector <string> localse;
  vector <Environment> middleware;
  int totaljobs;
  int usedcpus;
  int queuedjobs;
  string location;
  vector <string> owner;
  string issuer;
  vector <string> nodeaccess;
  LdapQuery ldap;
 public:
  vector <Queue> queues;
};


void FindClusterInfo (vector <Cluster> & clusterlist,
		      const string & usersn,
		      bool anonymous,
		      int timeout,
		      int debug = 0);


vector <Cluster> FindClusters (vector <Giis> giislist,
			       const string & usersn,
			       bool anonymous,
			       int timeout,
			       int debug = 0);

#endif
