#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <vector>
#include <map>

#ifndef UI_REMOTEQUERY
#define UI_REMOTEQUERY

class Cluster;
class Queue;


enum FileLocation {
  REMOTE,
  LOCAL,
  REMOTENOCACHE,
  LOCALNOCACHE,
  CACHED,
  NOLOCATION,
  IOERROR
};


class RemoteFileInfo {

 public:
  RemoteFileInfo (const char T);
  virtual ~RemoteFileInfo () {};
  virtual bool operator== (const RemoteFileInfo & file) const = NULL;
  virtual bool IsLocal (Cluster * c) const;
  virtual int Query () = NULL;
  long long int GetSize ();
  bool IsType (const char T) const;
  void RegisterCachedFile (const string & cluster);
  bool IsCached (const string & cluster) const;
 protected:
  long long int size;
  bool queried;
 private:
  const char type;
  vector <string> cached;
};


class RcLocationInfo {

 public:
  RcLocationInfo (const string & name, const string & uc, const string & path);
  ~RcLocationInfo () {};
  const string & GetName () const;
  const string & GetUrl () const;
  const string   GetHost () const;
 private:
  string name;
  string url;
};


class RcFileInfo : public RemoteFileInfo {

 public:
  RcFileInfo (const string & rc, const string & lfn);
  virtual ~RcFileInfo () {};
  virtual bool operator== (const RemoteFileInfo & file) const;
  virtual int Query ();
 private:
  const string rc;
  const string lfn;
 public:
  vector <RcLocationInfo> loclist;
};


#ifdef HAVE_GLOBUS_RLS_CLIENT_H
class RlsLocationInfo {

 public:
  RlsLocationInfo (const string & url);
  ~RlsLocationInfo () {};
  const string & GetUrl () const;
  const string   GetHost () const;
 private:
  string url;
};


class RlsFileInfo : public RemoteFileInfo {

 public:
  RlsFileInfo (const string & server, const string & lfn);
  virtual ~RlsFileInfo () {};
  virtual bool operator== (const RemoteFileInfo & file) const;
  virtual int Query ();
 private:
  const string server;
  const string lfn;
 public:
  vector <RlsLocationInfo> loclist;
};
#endif


class FtpFileInfo : public RemoteFileInfo {

 public:
  FtpFileInfo (const string & url);
  virtual ~FtpFileInfo () {};
  virtual bool operator== (const RemoteFileInfo & file) const;
  virtual bool IsLocal (Cluster * c) const;
  virtual int Query ();
 private:
  const string url;
};


class HttpFileInfo : public RemoteFileInfo {

 public:
  HttpFileInfo (const string & url);
  virtual ~HttpFileInfo () {};
  virtual bool operator== (const RemoteFileInfo & file) const;
  virtual bool IsLocal (Cluster * c) const;
  virtual int Query ();
 private:
  const string url;
};


class LocalFileInfo : public RemoteFileInfo {

 public:
  LocalFileInfo (const string & filename);
  virtual ~LocalFileInfo () {};
  virtual bool operator== (const RemoteFileInfo & file) const;
  virtual int Query ();
 private:
  const string filename;
};


class RemoteFile {

 public:
  RemoteFile (const string & optstring);
  virtual ~RemoteFile () {};
  virtual FileLocation Local (Cluster * c) const;
  long long int GetSize () const;
  bool IsType (const char T) const;
  void RegisterCachedFile (const string & cluster) const;
 protected:
  void Register ();
  RemoteFileInfo * fileinfo;
  map <string, string> options;
 private:
  static vector <RemoteFileInfo *> filecloud;
};


class RcLocation {

 public:
  RcLocation (const string & name, const string & optstring);
  ~RcLocation () {};
  const string & GetName () const;
  bool GetBooleanOption (const string & opt, bool * value) const;
 private:
  string name;
  map <string, string> options;
};


class RcFile : public RemoteFile {

 public:
  RcFile (const string & rc, const string & lfn,
	  const string & locstring, const string & optstring);
  virtual ~RcFile () {};
  virtual FileLocation Local (Cluster * c) const;
 private:
  vector <RcLocation> loclist;
  map <string, string> defaultoptions;
};


#ifdef HAVE_GLOBUS_RLS_CLIENT_H
class RlsLocation {

 public:
  RlsLocation (const string & url, const string & optstring);
  ~RlsLocation () {};
  const string & GetUrl () const;
  bool GetBooleanOption (const string & opt, bool * value) const;
 private:
  string url;
  map <string, string> options;
};


class RlsFile : public RemoteFile {

 public:
  RlsFile (const string & server, const string & lfn,
	   const string & locstring, const string & optstring);
  virtual ~RlsFile () {};
  virtual FileLocation Local (Cluster * c) const;
 private:
  vector <RlsLocation> loclist;
  map <string, string> defaultoptions;
};
#endif


class FtpFile : public RemoteFile {

 public:
  FtpFile (const string & url, const string & optstring);
  virtual ~FtpFile () {};
};


class HttpFile : public RemoteFile {

 public:
  HttpFile (const string & url, const string & optstring);
  virtual ~HttpFile () {};
};


class LocalFile : public RemoteFile {

 public:
  LocalFile (const string & filename, const string & optstring);
  virtual ~LocalFile () {};
  virtual FileLocation Local (Cluster * c) const;
};


class RemoteFileQuery {

 public:
  RemoteFileQuery () {};
  RemoteFileQuery (const RemoteFileQuery & query);
  ~RemoteFileQuery ();
  RemoteFileQuery & operator= (const RemoteFileQuery & query);
  void AddFile (RemoteFile * const file);
  int AddSizes (Cluster * c,
		long long int * cachesize,
		long long int * sessdirsize,
		long long int * remotesize,
		long long int * localsize) const;
  int RegisterCachedFiles (Cluster * c) const;
 private:
  vector <RemoteFile *> filelist;
};


#endif
